Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import ssralg fintype.
Require Import dvdring matrix mxalgebra bigop zmodp perm.
Require Import seqmatrix.

Section smith_def.

Local Open Scope ring_scope.

Variable R : euclidRingType.

Local Notation "''M_' ( m , n )" := 'M[R]_(m, n) : type_scope.
Local Notation "''M_' ( n )" := 'M[R]_(n, n) : type_scope.
Local Notation "''M_' n" := 'M[R]_(n, n) : type_scope.



Definition combine_mx (a b c d : R) (m : nat) (k : 'I_m) :=
  let k' := lift 0 k in
  let d := \row_j (a *+ (j == 0) + d *+ (j == k') + ((j != 0) && (j != k'))%:R) in
  diag_mx d + c *: delta_mx k' 0 + b *: delta_mx 0 k'.

(*
Definition bz_mx_inv (a b : R) (m : nat) (k : 'I_m) :=
  let:(g,u,v,a1,b1) := egcdr a b in
  let k' := lift 0 k in
  let d := \row_j (a1 *+ (j == 0) + u *+ (j == k') + ((j != 0) && (j != k'))%:R) in
  diag_mx d + b1 *: delta_mx k' 0 - v *: delta_mx 0 k'.
*)

(*
Definition bz_mx (a b : R) (m : nat) (k : 'I_m) : 'M_(1+m) :=
  let:(g,u,v,a1,b1) := egcdr a b in
  let d := \row_(j < m) (1 *+ (j != k) + a1 *+ (j == k)) in
  block_mx (a%:M) (-b1 *: delta_mx 0 k) (v *: delta_mx k 0) (diag_mx d).
*)



(* Local Close Scope ring_scope. *)

Import GRing.Theory.

Lemma combine_mxE (a b c d : R) (m n : nat) (M : 'M_(1 + m,1 + n)) (k : 'I_m) :
  let k' := lift 0 k in
  let r0 := a *: row 0 M + b *: row k' M in
  let rk := c *: row 0 M + d *: row k' M in
  combine_mx a b c d m k *m M =
    \matrix_i (r0 *+ (i == 0) + rk *+ (i == k') + row i M *+ ((i != 0) && (i != k'))).
Proof.
rewrite /combine_mx; apply/matrixP=> i j.
rewrite !mxE (bigD1 (0 : 'I_m.+1)) // !mxE (bigD1 (lift 0 k)) // !mxE /=.
case H:(i==0).
  rewrite big1=> [|l /andP [H1 H2]].
    rewrite (eqP H) !eqxx !mulr1n !mxE !mulr0 !addr0.
    by rewrite !mulr1 mulr0n add0r.
  rewrite !mxE (eqP H) (eq_sym 0 l) (negbTE H1) (negbTE H2).
  by rewrite mulr0n !mulr0 !add0r mul0r.
case H':(i==lift 0 k).
  rewrite big1=> [|l /andP [H1 H2]].
    rewrite (eqP H') !eqxx !mulr1n !mxE !mulr0 !addr0.
    by rewrite !mulr1 mulr0n !add0r addrC.
  rewrite !mxE (eqP H') !(eq_sym (lift 0 k) l) eqxx (negbTE H1) (negbTE H2).
  by rewrite mulr0n !mulr0 !add0r mul0r.
rewrite (bigD1 i); last by rewrite H H'.
rewrite !mxE big1=> [/=|l /andP [/andP [H1 H2] H3]].
  rewrite H H' eqxx !mulr0n !mulr0 !mulr1n.
  by rewrite !addr0 !mul0r !add0r mul1r.
rewrite !mxE H H' (negbTE H1) (negbTE H2) (eq_sym i l) (negbTE H3) /=.
by rewrite mulr0n !mulr0 !addr0 mul0r.
Qed.

(* Try to rewrite without case analysis *)
Lemma combine_mx_inv (a b c d : R) m k :
  a * d - b * c = 1 ->
  combine_mx a b c d m k *m combine_mx d (-b) (-c) a m k= 1%:M.
Proof.
move=> H; rewrite combine_mxE; apply/matrixP=> i j; rewrite !mxE.
case Hi:(i==0).
  rewrite !mxE (eqP Hi) !eqxx !mulr0 mxE !addr0 (eq_sym 0 j).
  case Hj:(j==0).
    rewrite (eqP Hj) mulr1 !mulr0.
    by rewrite addr0 sub0r mulrN.
  rewrite !mulr0 !add0r addr0 (eq_sym _ j).
  case:(j==lift 0 k); last by rewrite !mulr0 add0r.
  by rewrite mulr1 mulr1n mulrN mulrC addNr.
case Hj:(j==0).
  rewrite !mxE (eqP Hj) Hi add0r.
  case Hk:(i==lift 0 k).
    rewrite !mxE !eqxx !mulr0 mulr1 !addr0 !add0r.
    by rewrite mulrN addrC mulrC addNr.
  by rewrite !mxE Hi Hk eqxx !add0r !mulr0 addr0.
case Hk:(i==lift 0 k).
  rewrite !mxE (eq_sym 0 j) Hj (eqP Hk) !eqxx !mulr0 !addr0 !add0r.
  rewrite (eq_sym _ j).
  case:(j==lift 0 k); last by rewrite !mulr0 addr0.
  by rewrite !mulr1 addrC mulrN (mulrC c) (mulrC d).
by rewrite !mxE Hi Hj Hk !mulr0 !add0r !addr0.
Qed.


(*****************
  if the following Bezout identity holds: u * a1 + v * b1 = 1,
  bz_mx a b n k represents the following matrix (dots are zeros):

          (kth column)
  / u .... v ..... \
  | . 1 .......... |
  | ....1......... |
  | -b1 .. a1 .... | (kth row)
  | ..........1... |
  \ .............1 /


  (determinant is +/-1)
******************)

Definition bz_mx (a b : R) (m : nat) (k : 'I_m) : 'M_(1+m) :=
  let:(_,u,v,a1,b1) := egcdr a b in combine_mx u v (-b1) a1 m k.

Lemma bz_mx_00 {m n : nat} {M : 'M_(1 + m,1 + n)} {k : 'I_m} :
 (bz_mx (M 0 0) (M (lift 0 k) 0) m k *m M) 0 0 %= gcdr (M 0 0) (M (lift 0 k) 0).
Proof.
rewrite /bz_mx; case:egcdrP=> g u v a1 b1 Hg Hgcd H1 H2.
by rewrite combine_mxE !mxE !addr0 {1}H1 {1}H2 !mulrA -mulr_addl Hg mul1r.
Qed.

Lemma sdvd_bz_mx (m n : nat) (M : 'M_(1 + m,1 + n)) (k : 'I_m) :
 ~~ (M 0 0 %| M (lift 0 k) 0) ->
 (bz_mx (M 0 0) (M (lift 0 k) 0) m k *m M) 0 0 %<| M 0 0.
Proof.
move=> H ; rewrite /sdvdr (eqd_dvd (bz_mx_00) (eqdd _)) dvdr_gcdl.
rewrite (eqd_dvd (eqdd _ ) (bz_mx_00)).
by apply/negP=> H'; rewrite (dvdr_trans H' (dvdr_gcdr _ _)) in H.
Qed.

(*
Lemma bz_mx_invP (a b : R) m k : bz_mx a b m k *m bz_mx_inv a b m k = 1%:M.
Proof.
rewrite bz_mxE /bz_mx_inv.
apply/matrixP=> i j; rewrite !mxE.
case:egcdrP=> g u v a1 b1 Huv Hg Ha Hb /=.
case Hi:(i==0).
  case Hj:(j==0).
    rewrite !mxE (eqP Hi) (eqP Hj) !mulr1n !mulr0n !eqxx.
    rewrite mxE !mulr0 !addr0 add0r !subr0 mulr1.
    admit.
  rewrite !mxE (eqP Hi) !mulr1n !mulr0n !eqxx.
  rewrite mxE !mulr0 !addr0 add0r !subr0 mulr1.

rewrite /bz_mx /bz_mx_inv.
*)

Lemma unit_bz_mx m n (M : 'M_(1 + m,1 + n)) k :
  (bz_mx (M 0 0) (M (lift 0 k) 0) m k) \in unitmx.
Proof.
rewrite /bz_mx; case:egcdrP=> g a1 b1 u v Huv Hg Ha1 Hb1.
have H: a1 * u - b1 * -v = 1; first by rewrite mulrN opprK.
by case:(mulmx1_unit (combine_mx_inv _ _ _ _ m k H)).
Qed.

Local Open Scope ring_scope.

Fixpoint improve_pivot k {m n} : 'M_(1 + m, 1 + n) -> 'M_(1 + m) * 'M_(1 + m, 1 + n) * 'M_(1 + n) :=
  match k with
  | 0 => fun A => (1%:M,A,1%:M)
  | p.+1 => fun A =>
    let a := A 0 0 in
    if [pick i | ~~(a %| A (lift 0 i) 0)] is Some i then
        let L := bz_mx (A 0 0) (A (lift 0 i) 0) m i in
        let A := L *m A in
        let:(L',A,R') := improve_pivot p A in
        (L'*m L, A, R')
    else if [pick j | ~~(a %| A 0 (lift 0 j))] is Some j then
        let R := (bz_mx (A 0 0) (A 0 (lift 0 j)) n j)^T in
        let A := A *m R in
        let:(L,A,R') := improve_pivot p A in
        (L, A, R *m R')
    else (1%:M,A,1%:M)
  end.

Section improve_pivot_correct.

Variables m n : nat.

CoInductive improve_pivot_spec M :
  'M_(1 + m) * 'M_(1 + m,1 + n) * 'M_(1 + n) -> Type :=
  ImprovePivotSpec L A R of L *m M *m R = (A : 'M_(1 + m, 1 + n))
  & (forall i, A 0 0 %| A i 0) & (forall j, A 0 0 %| A 0 j)
  & A 0 0 %| M 0 0
  & L \in unitmx & R \in unitmx
  : improve_pivot_spec M (L,A,R).

Lemma improve_pivotP : forall k (M : 'M_(1 + m,1 + n)),
  (enorm (M 0%R 0%R) <= k)%N -> M 0 0 != 0 ->
    improve_pivot_spec M (improve_pivot k M).
Proof.
elim=> [M|k IHk M Hk H /=].
  by rewrite leqn0 ; move/eqP/enorm_eq0 -> ; rewrite eqxx.
case:pickP=> [i Hi|Hi].
  case:(IHk (_ *m M))=> [||L' A' R'].
  + by rewrite -ltnS; apply:(leq_trans (ltn_enorm H (sdvd_bz_mx _ _ _ _ Hi)) Hk).
  + by rewrite -eqdr0 (congr_eqd bz_mx_00 (eqdd _)) eqdr0 gcdr_eq0 (negbTE H).
  + rewrite mulmxA=> HA' ? ? Hdiv HL' HR'; constructor=> //.
      rewrite (eqd_dvd (eqdd _) bz_mx_00) in Hdiv.
      exact:(dvdr_trans Hdiv (dvdr_gcdl _ _)).
    by rewrite unitmx_mul HL' unit_bz_mx.
case:pickP=> [j|Hj].
  move:H Hk; rewrite -(trmxK M) mxE [_ 0 (lift 0 j)]mxE -trmx_mul=> H Hk Hj.
  case:IHk=> [||L' A' R'].
  + by rewrite -ltnS mxE; apply:(leq_trans (ltn_enorm H (sdvd_bz_mx _ _ _ _ Hj)) Hk).
  + by rewrite mxE -eqdr0 (congr_eqd bz_mx_00 (eqdd _)) eqdr0 gcdr_eq0 (negbTE H).
  + rewrite {1}trmx_mul mulmxA=> HA' ? ? Hdiv HL' HR'; constructor; rewrite ?mulmxA //.
      rewrite mxE (eqd_dvd (eqdd _) bz_mx_00) in Hdiv.
      by rewrite mxE; exact:(dvdr_trans Hdiv (dvdr_gcdl _ _)).
    by rewrite unitmx_mul HR' unitmx_tr unit_bz_mx.
constructor=> [|i|j|||]; first by rewrite mul1mx mulmx1.
  + case:(@splitP 1 m i)=> i'.
      by rewrite (ord1 i')=> Hi'; rewrite (@ord_inj _ i 0 Hi').
    rewrite (add1n i') -lift0 => Hi'; rewrite (ord_inj Hi').
    exact:(negbFE (Hi i')).
  + case:(splitP j)=> j'.
      by rewrite (ord1 j')=> Hj'; rewrite (@ord_inj _ j 0 Hj').
    rewrite (add1n j') -lift0 => Hj'; rewrite (ord_inj Hj').
    exact:(negbFE (Hj j')).
  + exact:dvdrr.
  + exact:unitmx1.
  + exact:unitmx1.
Qed.

End improve_pivot_correct.

(* ... max m.+1 n.+1 -> (max m n).+1 *)

(*
Lemma widen_proof m n : (m <= maxn m n)%N.
Proof. by rewrite leq_maxr leqnn. Qed.

Definition rect_diag_mx {m n} (d : 'rV[R]_(maxn m n)) :=
  \matrix_(i < m, j < n) (d 0 (widen_ord (widen_proof _ _) i) *+ (i == j :> nat)).

Lemma rect_diag_mx_nn n (d : 'rV[R]_(maxn n n)) :
  rect_diag_mx d = diag_mx (castmx (refl_equal _,maxnn n) d).
Proof.
apply/matrixP=> i j; rewrite !mxE; case:(i == j)/eqP=> [->|Hij].
  by rewrite eqxx castmxE cast_ord_id; congr fun_of_matrix; exact:val_inj.
case H:(i == j :> nat); first by move:(ord_inj (eqP H)).
by rewrite !mulr0n.
Qed.
*)

Fixpoint half_smith {m n} : 'M_(m,n) ->
    'M_(m) * 'rV_n * 'M_(n) :=
  match m, n return 'M_(m, n) -> 'M_(m) * 'rV_n * 'M_(n) with
  | _.+1, _.+1 => fun A : 'M_(1 + _, 1 + _) =>
      if [pick ij | A ij.1 ij.2 != 0] is Some (i, j) then
      let a := A i j in let A := xrow i 0 (xcol j 0 A) in
      (* this is where Euclidean norm eases termination argument *)
      let:(L,A,R) := improve_pivot (enorm a) A in
      let a := A 0 0 in
      let u := dlsubmx A in let v := ursubmx A in
      let u' := map_mx (fun x => odflt 0 (x %/? a)) u in
      let v' := map_mx (fun x => odflt 0 (x %/? a)) v in
      let A := (drsubmx A) - u' *m v (* - u *m v' *) in
      let:(L',A,R') := half_smith A in
      (lift0_mx L' *m block_mx 1 0 (-u') 1%:M *m (xcol i 0 L),
       row_mx a%:M A,
       (xrow j 0 R) *m block_mx 1 (-v') 0 1%:M *m lift0_mx R')
    else (1%:M, 0, 1%:M)
  | _, _ => fun A => (1%:M, 0, 1%:M)
  end.

CoInductive half_smith_spec {m n} M : 'M[R]_m * 'rV[R]_n * 'M[R]_n -> Type :=
  HalfSmithSpec L d R of L *m M *m R = (@pid_mx _ m n m *m (diag_mx d))
  & (forall i, d 0 i = 0 -> forall (j : 'I_n), (i <= j)%N -> d 0 j = 0)
  & L \in unitmx & R \in unitmx : half_smith_spec M (L,d,R).

Lemma unit_lift0_mx (n:nat) (M: 'M[R]_n) :
 M \in unitmx = (lift0_mx M \in unitmx).
Proof. by rewrite !unitmxE det_ublock det1 mul1r. Qed.

Lemma diag_row_mx m (a : R) (A:'rV_m) :
  diag_mx (row_mx a%:M A) = block_mx a%:M 0 0 (diag_mx A).
Proof.
apply/matrixP=> i j; rewrite !mxE eqE /=.
case: splitP=> i' ->; rewrite !mxE; case: splitP=> j' ->;
by rewrite ?(ord1 i') ?(ord1 j') ?eqxx mxE.
Qed.

Lemma pid_mx_block_1 m n :
  pid_mx (1 + m) = block_mx 1%:M 0 0 (pid_mx m) :> 'M_(1 + m, 1 + n).
Proof.
apply/matrixP=> i j; rewrite !mxE.
case: splitP => i' ->; rewrite mxE; case: splitP => j' ->; rewrite mxE //;
by rewrite ?(ord1 i') ?(ord1 j') ?eqxx.
Qed.

Lemma half_smithP : forall (m n : nat) (M : 'M_(m,n)),
  half_smith_spec M (half_smith M).
Proof.
elim=> [n M|m IHn]; first constructor=> [|i _ j _||]; rewrite ?mxE ?unitmx1 //.
  by rewrite [M]flatmx0 [pid_mx _]flatmx0 diag_const_mx mulmx0 !mul0mx.
case=> [M|n M]; first constructor=> [|i _ j _||]; rewrite ?mxE ?unitmx1 //.
  by rewrite [M]thinmx0 [pid_mx _]thinmx0 mulmx0 !mul0mx.
rewrite /half_smith -{1}/half_smith.
case:pickP=>[[i j] HMij | H].
  case:improve_pivotP=> [||L A R0 HA HAi0 HA0j Hdiv]; rewrite ?mxE ?tpermR ?leqnn //.
  set V := map_mx _ _; set U := map_mx _ _.
  case:IHn=> L' A' R' HA' H0 HL' HR' HL HR.
  constructor.
  * rewrite xcolE xrowE -!mulmxA (mulmxA M) -xcolE (mulmxA (tperm_mx _ _)).
    rewrite -xrowE (mulmxA L) (mulmxA _ R0) HA mulmx_block !mulmxA mulmx_block.
    rewrite -{1}(submxK A) !mulmx_block.
    do 2! rewrite !mul0mx !mulmx0 !mulmx1 !mul1mx !addr0 ?add0r.
    have HV: (V *m ulsubmx A) = dlsubmx A.
      apply/matrixP=> k l.
      rewrite (ord1 l) !mxE big_ord_recl big_ord0 !mxE !lshift0 addr0.
      case:odivrP=>[x ->|H] //; case/dvdrP:(HAi0 (rshift 1 k))=> q.
      by move/eqP; rewrite (negbTE (H q)).
    have HU: (ulsubmx A *m U) = ursubmx A.
      apply/matrixP=> k l.
      rewrite (ord1 k) !mxE big_ord_recl big_ord0 !mxE !lshift0 addr0.
      case:odivrP=>[x ->|H]; first by rewrite mulrC.
      by case/dvdrP:(HA0j (rshift 1 l))=> q; move/eqP; rewrite (negbTE (H q)).
    rewrite diag_row_mx pid_mx_block_1 mulmx_block !mulmx0 !mul0mx addr0 !add0r.
    congr block_mx.
    (* Pivot *)
    + apply/matrixP=> k l; rewrite (ord1 k) (ord1 l).
      by rewrite !mxE big_ord1 !mxE eqxx !lshift0 mul1r.
    (* Horizontal zeros *)
    + by rewrite mulNmx mulmxN mulmxA HU addNr.
    (* Vertical zeros *)
    + by rewrite -HV mulmxN mulNmx mulmxA addNr.
    (* down-right submatrix *)
    + rewrite mulmxN !mulNmx -mulmxA HV addNr mul0mx.
    by rewrite add0r addrC -mulmxA -mulmx_subr.
  * move=> k; rewrite mxE; case:(splitP k)=> k' ->.
      rewrite [k']ord1 mxE eqxx; have: ~(A 0 0 = 0) => // Habs.
      by rewrite Habs dvd0r !mxE !tpermR (negbTE HMij) in Hdiv.
    move=> Hk l; rewrite mxE; case:(splitP l)=> l' ->.
      by rewrite (ord1 l').
    exact:H0.
  * rewrite xcolE !unitmx_mul unitmx_perm HL !unitmxE.
    by rewrite !det_lblock !det1 mul1r mulr1 unitr1 -unitmxE !andbT.
  * rewrite xrowE !unitmx_mul unitmx_perm HR !unitmxE.
  by rewrite !det_ublock !det1 !mul1r unitr1.
constructor=>[|i _ j||]; rewrite ?mxE ?unitmx1 //.
rewrite mul1mx mulmx1 mul_mx_diag; apply/matrixP=> i j.
by rewrite !mxE mulr0 (eqP (negbFE (H (i,j)))).
Qed.


End smith_def.
