Require Import String.

Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import ssralg fintype perm.
Require Import dvdring matrix  bigop zmodp mxalgebra.
Require Import rank prime.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* to use the F2 implementation of SSR, we first need to prove
   that 2 is a prime number *)
Lemma p2 : prime 2%N.
Proof.
apply/primeP.
split => // d.
case : d => [ | d].
- by rewrite dvd0n.
move/dvdn_leq.
case: d => [ | d].
- by rewrite eq_refl.
case: d => [ | d] hd.
- by rewrite eq_refl orbT.
assert (0 < 2) by apply ltn0Sn.
have : d.+2 < 2 by apply hd.
change (d.+2) with (1 + (1 + d)).
change 2 with (1 + (1 + 0)).
by rewrite 2!ltn_add2l ltn0.
Qed.

Definition F2 := Fp_fieldType 2.

Definition e2rank := @ex_Rank F2.

Notation "n %:F2" := (n%R : 'F_2) (at level 2, left associativity, format "n %:F2").

(*
  0  1
  1  0
*)
Definition foo : Matrix F2 := cM 0 [::(1 %:F2)] [::1 %:F2 ] (cM 0 [::] [::] (@eM _)).

Eval compute in (e2rank 2 foo).


Definition ex1 := cM 1%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 1%:F2] [:: 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 1%:F2] [:: 0%:F2] (cM 0%:F2 [::] [::] (@eM _ ))))).
Definition ex1_size := 5.

Compute "ex1"%string.
Time Eval vm_compute in (ex_Rank ex1_size ex1).

Definition ex2 := cM 1%:F2 [:: 0%:F2] [:: 1%:F2] (cM 1%:F2 [::] [::] (@eM _ )).
Definition ex2_size := 2.

Compute "ex2"%string.
Time Eval vm_compute in (ex_Rank ex2_size ex2).


Definition ex3 := cM 1%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2] [:: 0%:F2] (cM 0%:F2 [::] [::] (@eM _ )))))))))))))))))))))))))))))))).
Definition ex3_size := 32.


Compute "ex3"%string.
Time Eval vm_compute in (ex_Rank ex3_size ex3).


Definition ex4 := cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2] [:: 0%:F2] (cM 0%:F2 [::] [::] (@eM _ )))))))))))))))).
Definition ex4_size := 16.

Compute "ex4"%string.
Time Eval vm_compute in (ex_Rank ex4_size ex4).


Definition ex5 := cM 1%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2] [:: 0%:F2] (cM 0%:F2 [::] [::] (@eM _ )))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))).
Definition ex5_size := 64.


Compute "ex5"%string.
Time Eval vm_compute in (ex_Rank ex5_size ex5).

Definition ex6 := cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 1%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 1%:F2 ; 1%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2] [:: 0%:F2] (cM 0%:F2 [::] [::] (@eM _ )))))))))))))))))))))))))))))))).
Definition ex6_size := 32.

Compute "ex6"%string.
Time Eval vm_compute in (ex_Rank ex6_size ex6).

Definition ex7 := cM 1%:F2 [:: 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 1%:F2 ; 1%:F2 ; 0%:F2] [:: 0%:F2 ; 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 0%:F2 ; 1%:F2] [:: 0%:F2] (cM 0%:F2 [:: 0%:F2 ; 1%:F2 ; 1%:F2] [:: ] ((@eM _ )))))))).
Definition ex7_size := 7.

Compute "ex7"%string.
Time Eval vm_compute in (ex_Rank ex7_size ex7).

Definition ex8 := cM 1%:F2 [:: 0%:F2] [:: 1%:F2 ; 0%:F2 ; 1%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2 ; 0%:F2] ((@eM _ )).
Definition ex8_size := 10.

Compute "ex8"%string.
Time Eval vm_compute in (ex_Rank ex8_size ex8).
